// SPDX-License-Identifier: BSD-3-Clause

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>

#include "os_graph.h"
#include "os_threadpool.h"
#include "log/log.h"
#include "utils.h"

#define NUM_THREADS		4

static int sum;
static os_graph_t *graph;
static os_threadpool_t *tp;
/* TODO: Define graph synchronization mechanisms. */

/* TODO: Define graph task argument. */
typedef struct {
	unsigned int x;
	void (*auxFunc)(unsigned int idx);
} os_arg_t;

static void destroy_arg(void *arg)
{
	free(arg);
}

static void action(void *arg)
{
	((os_arg_t *)arg)->auxFunc(((os_arg_t *)arg)->x);
}

static void process_node(unsigned int idx)
{
	/* TODO: Implement thread-pool based processing of graph. */
	pthread_mutex_lock(&tp->mutexCoada);
	sum += graph->nodes[idx]->info;
	graph->visited[idx] = DONE;
	for (unsigned int i = 0; i < graph->nodes[idx]->num_neighbours; ++i) {

		if (graph->visited[graph->nodes[idx]->neighbours[i]] == NOT_VISITED) {
			graph->visited[graph->nodes[idx]->neighbours[i]] = PROCESSING;
			os_arg_t *arg = malloc(sizeof(*arg));

			arg->x = graph->nodes[idx]->neighbours[i];
			arg->auxFunc = process_node;
			os_task_t *t = create_task(&action, (void *)arg, &destroy_arg);

			enqueue_task(tp, t);
		}
	}
	pthread_mutex_unlock(&tp->mutexCoada);
}

int verifProcessing(void)
{
	for (unsigned int i = 0; i < graph->num_nodes; ++i) {
		if (graph->visited[i] == PROCESSING)
			return 1;
	}
	return 0;
}

int main(int argc, char *argv[])
{
	FILE *input_file;

	if (argc != 2) {
		fprintf(stderr, "Usage: %s input_file\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	input_file = fopen(argv[1], "r");
	DIE(input_file == NULL, "fopen");

	graph = create_graph_from_file(input_file);

	/* TODO: Initialize graph synchronization mechanisms. */
	tp = create_threadpool(NUM_THREADS);
	process_node(0);
	tp->verifProcessing = verifProcessing;
	destroy_threadpool(tp);

	printf("%d", sum);

	return 0;
}
